# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""File transfers to upload and download files to / from a file server.

All file transfers are designed to work with :class:`scitacean.Client`.
Refer to the submodules for concrete setup and usage instructions.
"""
